# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Test ICAT client
This performs a large amount of mocking of the actual ICAT functions.
This is because we can not connect to icat for testing and it is not feasible
to set up a local version for testing at this point.
"""
import unittest
from unittest.mock import patch

import icat

from autoreduce_utils.clients.icat_client import ICATClient
from autoreduce_utils.clients.settings.client_settings_factory import ClientSettingsFactory
from autoreduce_utils.clients.connection_exception import ConnectionException


def raise_icat_session_error():
    """ function required to raise ICATSessionErrors in mocks """
    raise icat.ICATSessionError('ICAT session error raised from mock')


# pylint:disable=missing-docstring,no-self-use
class TestICATClient(unittest.TestCase):
    @patch('icat.Client.__init__', return_value=None)
    def test_default_init(
        self,
        mock_icat,
    ):
        """
        Test: Class variables are created and set
        When: ICATClient is initialised with default credentials
        """
        client = ICATClient()
        self.assertEqual(client.credentials.username, 'YOUR-ICAT-USERNAME')
        self.assertEqual(client.credentials.password, 'YOUR-PASSWORD')
        self.assertEqual(client.credentials.host, 'YOUR-ICAT-WSDL-URL')
        self.assertEqual(client.credentials.port, '')
        self.assertEqual(client.credentials.auth, 'simple')
        mock_icat.assert_called_once_with('YOUR-ICAT-WSDL-URL')

    @patch('icat.Client.__init__', return_value=None)
    @patch('icat.Client.login')
    def test_valid_connection(self, mock_icat_login, mock_icat):
        """
        Test: login is called on the stored client
        When: connect is called while valid credentials are held
        """
        client = ICATClient()
        mock_icat.assert_called_once()
        client.connect()
        mock_icat_login.assert_called_once_with(auth='simple',
                                                credentials={
                                                    'username': 'YOUR-ICAT-USERNAME',
                                                    'password': 'YOUR-PASSWORD'
                                                })

    def test_invalid_connection(self):
        """
        Test: A ValueError is raised
        When: connect is called while invalid credentials are held
        """
        invalid_settings = ClientSettingsFactory().create('icat',
                                                          username='user',
                                                          password='pass',
                                                          host=r'www.fake-url.com',
                                                          port='',
                                                          authentication_type='none')
        self.assertRaises(ValueError, ICATClient, invalid_settings)

    @patch('icat.Client.__init__', return_value=None)
    @patch('icat.Client.logout')
    def test_disconnect(self, mock_logout, _):
        """
        Test: logout is called on the stored client
        When: disconnect is called
        """
        client = ICATClient()
        client.disconnect()
        mock_logout.assert_called_once()

    @patch('icat.Client.__init__', return_value=None)
    @patch('icat.Client.refresh')
    def test_refresh(self, mock_autorefresh, _):
        """
        Test: refresh is called on the stored client
        When: refresh is called
        """
        client = ICATClient()
        client.refresh()
        mock_autorefresh.assert_called_once()

    @patch('icat.Client.__init__', return_value=None)
    @patch('icat.Client.refresh')
    def test_valid_test_connection(self, mock_refresh, _):
        """
        Test: refresh is called on the stored client
        When: _test_connection is called
        """
        client = ICATClient()
        # pylint:disable=protected-access
        self.assertTrue(client._test_connection())
        mock_refresh.assert_called_once()

    @patch('icat.Client.__init__', return_value=None)
    @patch('icat.Client.refresh')
    def test_failing_test_connection(self, mock_refresh, _):
        """
        Test: A ConnectionException is raised
        When: _test_connection is called while an invalid connection is held
        """
        mock_refresh.side_effect = raise_icat_session_error
        client = ICATClient()
        # pylint:disable=protected-access
        self.assertRaisesRegex(ConnectionException, 'ICAT', client._test_connection)

    @patch('icat.Client.__init__', return_value=None)
    @patch('icat.Client.__setattr__')
    @patch('autoreduce_utils.clients.icat_client.ICATClient.connect')
    @patch('icat.Client.refresh')
    @patch('icat.Client.search')
    def test_query_without_conn(self, mock_search, mock_refresh, mock_connect, mock_set_attr, _):
        """
        Test: A query is executed
        When: execute_query is called without a connection having been established
        """
        # Add side effect to raise exception with icat.refresh
        mock_refresh.side_effect = raise_icat_session_error
        client = ICATClient()
        client.execute_query('icat query')
        mock_refresh.assert_called_once()
        mock_set_attr.assert_called_once()
        mock_connect.assert_called_once()
        mock_search.assert_called_once_with('icat query')

    @patch('icat.Client.__init__', return_value=None)
    @patch('icat.Client.search')
    @patch('icat.Client.refresh', return_value=None)
    def test_query_icat(self, mock_refresh, mock_search, _):
        """
        Test: A query is executed
        When: execute_query is called with a connection having been established
        """
        client = ICATClient()
        client.execute_query('icat query')
        mock_refresh.assert_called_once()
        mock_search.assert_called_once_with('icat query')



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Client class for accessing queuing service
"""
import logging
import os
import uuid
import socket

import stomp
from stomp.exception import ConnectFailedException

from autoreduce_utils.message.message import Message
from autoreduce_utils.clients.abstract_client import AbstractClient
from autoreduce_utils.clients.connection_exception import ConnectionException
from autoreduce_utils.settings import ACTIVEMQ_SETTINGS


class QueueClient(AbstractClient):
    """
    Class for client to access messaging service via python
    """
    def __init__(self, credentials=None, consumer_name='queue_client'):
        if not credentials:
            credentials = ACTIVEMQ_SETTINGS
        super(QueueClient, self).__init__(credentials)  # pylint:disable=super-with-arguments
        self._connection = None
        self._consumer_name = consumer_name
        self._logger = logging.getLogger("queue_listener")

    def connect(self):
        """
        Create the connection if the connection has not been created
        """
        if self._connection is None or not self._connection.is_connected():
            self.disconnect()
            self._create_connection()

    def _test_connection(self):
        if self._connection.is_connected():
            return True
        raise ConnectionException("ActiveMQ")

    def disconnect(self):
        """
        Disconnect from queue service
        """
        self._logger.info("Starting disconnect from ActiveMQ...")
        if self._connection is not None and self._connection.is_connected():
            # By passing a receipt Stomp will call stop on the transport layer
            # which causes it to wait on the listener thread (if it's still
            # running). Without this step we just drop out, so the behaviour
            # is not guaranteed. UUID is used by Stomp if we don't pass in
            # a receipt so this matches the behaviour under the hood
            try:
                self._connection.remove_listener("queue_processor")
            except KeyError:
                pass  # If no listener was set for this client instance
            self._connection.disconnect(receipt=str(uuid.uuid4()))
            self._logger.info("Disconnected from ActiveMQ.")
        self._connection = None

    def _create_connection(self):
        """
        Create the connection to the queuing service and store as self._connection
        """
        inteded_for_production = "AUTOREDUCTION_PRODUCTION" in os.environ
        aimed_at_dev = self.credentials.host.startswith("127") or "dev" in str(
            self.credentials.host) or self.credentials.host == "activemq"
        # Prevent unintentional submission to non-development envs
        if not inteded_for_production and not aimed_at_dev:
            raise RuntimeError(
                f"Trying to submit to a potentially non-development environment at `{self.credentials.host}`. "
                "You must declare AUTOREDUCTION_PRODUCTION in the environment "
                "if you intend to submit to the production environment")
        if inteded_for_production and aimed_at_dev:
            raise RuntimeError(f"Trying to submit to production environment but host is `{self.credentials.host}`. "
                               "Remove AUTOREDUCTION_PRODUCTION if that is unintentional.")

        if self._connection is None or not self._connection.is_connected():
            try:
                host_port = [(self.credentials.host, int(self.credentials.port))]
                connection = stomp.Connection(host_and_ports=host_port, use_ssl=False)
                self._logger.info("Starting connection to %s", host_port)
                connection.connect(username=self.credentials.username, passcode=self.credentials.password, wait=True)
            except ConnectFailedException as exp:
                raise ConnectionException("ActiveMQ") from exp
            self._connection = connection

    def subscribe(self, listener):
        """
        Subscribes to the data_ready queue
        :param listener: QueueListener object
        """
        self._logger.info("Subscribing to data ready queue")
        self._connection.set_listener("queue_processor", listener)
        self._connection.subscribe(destination=self.credentials.data_ready,
                                   id=socket.getfqdn(),
                                   ack="client-individual",
                                   header={"activemq.prefetchSize": "1"})

    def ack(self, message_id, subscription):
        """
        Acknowledge receipt of a message
        :param message_id: The identifier of the message
        """
        # pylint:disable=no-value-for-parameter
        self._connection.ack(message_id, subscription)

    def send(self, destination, message, persistent='true', priority='4', delay=None):
        """
        Send a message via the open connection to a queue
        :param destination: Queue to send to
        :param message: Message instance OR json dump of dict containing message payload
        :param persistent: should to message be persistent
        :param priority: priority rating of the message
        :param delay: time to wait before send
        """
        self.connect()

        if isinstance(message, Message):
            message_json_dump = message.serialize()
        else:
            message_json_dump = message

        self._connection.send(destination, message_json_dump, persistent=persistent, priority=priority, delay=delay)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Test the ClientSettings package
"""
from unittest import TestCase, mock

from autoreduce_utils.clients.settings.client_settings import ClientSettings


# pylint:disable=missing-docstring
class TestClientSettings(TestCase):
    def test_valid_init(self):
        settings = ClientSettings(username='user', password='pass', host='host', port='123')
        self.assertIsNotNone(settings)
        self.assertEqual(settings.username, 'user')
        self.assertEqual(settings.password, 'pass')
        self.assertEqual(settings.host, 'host')
        self.assertEqual(settings.port, '123')

    def test_invalid_init(self):
        self.assertRaisesRegex(ValueError, "True of type <class 'bool'> is not a string", ClientSettings, 'string',
                               'string', True, 99.99)
        self.assertRaisesRegex(ValueError, "99.99 of type <class 'float'> is not a string", ClientSettings, 'string',
                               'string', 'string', 99.99)

    def test_accepts_mock_types(self):
        settings = ClientSettings(username=mock.Mock(),
                                  password=mock.MagicMock(),
                                  host=mock.NonCallableMock(),
                                  port=mock.Mock())

        self.assertIsNotNone(settings)
        self.assertIsInstance(settings.username, str)
        self.assertIsInstance(settings.password, str)
        self.assertIsInstance(settings.host, str)
        self.assertIsInstance(settings.port, str)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
# pylint: skip-file
"""
Settings for connecting to the test services that run locally
"""
import configparser
import os
from pathlib import Path

from autoreduce_utils.clients.settings.client_settings_factory import ClientSettingsFactory

CONFIG_ROOT = str(Path("~/.autoreduce").expanduser())
INI_FILE = os.environ.get("AUTOREDUCTION_CREDENTIALS", os.path.expanduser(f"{CONFIG_ROOT}/credentials.ini"))

CONFIG = configparser.ConfigParser()
CONFIG.read(INI_FILE)


def get_setting(section: str, key: str) -> str:
    """
    Gets the value of the key from the section.
    """
    return str(CONFIG.get(section, key, raw=True))  # raw=True to allow strings with special characters to be passed


SETTINGS_FACTORY = ClientSettingsFactory()

ICAT_SETTINGS = SETTINGS_FACTORY.create('icat',
                                        username=get_setting('ICAT', 'user'),
                                        password=get_setting('ICAT', 'password'),
                                        host=get_setting('ICAT', 'host'),
                                        port='',
                                        authentication_type=get_setting('ICAT', 'auth'))

MYSQL_SETTINGS = SETTINGS_FACTORY.create('database',
                                         username=get_setting('DATABASE', 'user'),
                                         password=get_setting('DATABASE', 'password'),
                                         host=get_setting('DATABASE', 'host'),
                                         port=get_setting('DATABASE', 'port'),
                                         database_name=get_setting('DATABASE', 'name'))

ACTIVEMQ_SETTINGS = SETTINGS_FACTORY.create('queue',
                                            username=get_setting('QUEUE', 'user'),
                                            password=get_setting('QUEUE', 'password'),
                                            host=get_setting('QUEUE', 'host'),
                                            port=get_setting('QUEUE', 'port'))

LOCAL_MYSQL_SETTINGS = SETTINGS_FACTORY.create('database',
                                               username=get_setting('DATABASE', 'user'),
                                               password=get_setting('DATABASE', 'password'),
                                               host=get_setting('DATABASE', 'host'),
                                               port=get_setting('DATABASE', 'port'))

SFTP_SETTINGS = SETTINGS_FACTORY.create('sftp',
                                        username=get_setting('SFTP', 'user'),
                                        password=get_setting('SFTP', 'password'),
                                        host=get_setting('SFTP', 'host'),
                                        port=get_setting('SFTP', 'port'))

CYCLE_SETTINGS = SETTINGS_FACTORY.create('cycle',
                                         username=get_setting('CYCLE', 'user'),
                                         password=get_setting('CYCLE', 'password'),
                                         host='',
                                         port='',
                                         uows_url=get_setting('CYCLE', 'uows_url'),
                                         scheduler_url=get_setting('CYCLE', 'scheduler_url'))



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Command to move test_settings.py to settings.py
"""
import sys
import os
import logging
from shutil import copyfile
from pathlib import Path

logger = logging.getLogger("autoreduce_utils")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))


def migrate_credentials(destination_path: str):
    """
    Migrate the credentials.ini
    """
    utils_path = os.path.join(ROOT_DIR, 'autoreduce_utils')
    logger.info("================== Migrate credentials ====================")
    test_credentials_path = os.path.join(utils_path, 'test_credentials.ini')
    cred_dir = Path(destination_path).expanduser()
    cred_dir.mkdir(parents=True, exist_ok=True)
    credentials_path = cred_dir / "credentials.ini"
    logger.info("Copying credentials to %s", credentials_path)
    copyfile(test_credentials_path, credentials_path)
    logger.info("Credentials successfully migrated")


def main():
    """
    Entrypoint for the migrate credentials action
    """
    destination_path = "~/.autoreduce/"
    migrate_credentials(destination_path)


if __name__ == "__main__":
    main()



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Represents the messages passed between AMQ queues
"""
import json

import attr

from autoreduce_utils.message.validation import stages


@attr.s
class Message:
    """
    A class that represents an AMQ Message.
    Messages can be serialized and deserialized for sending messages to and from AMQ
    """
    description = attr.ib(default="")
    facility = attr.ib(default="ISIS")
    run_number = attr.ib(default=None)
    instrument = attr.ib(default=None)
    rb_number = attr.ib(default=None)
    started_by = attr.ib(default=None)
    data = attr.ib(default=None)
    overwrite = attr.ib(default=None)
    run_version = attr.ib(default=None)
    job_id = attr.ib(default=None)
    reduction_script = attr.ib(default=None)
    reduction_arguments = attr.ib(default={})
    reduction_log = attr.ib(default="")  # Cannot be null in database
    admin_log = attr.ib(default="")  # Cannot be null in database
    message = attr.ib(default=None)
    retry_in = attr.ib(default=None)
    reduction_data = attr.ib(default=None)  # Required by reduction runner
    software = attr.ib(default=None)
    flat_output = attr.ib(default=False)

    def serialize(self, indent=None, limit_reduction_script=False):
        """
        Serialized member variables as a json dump
        :param indent: The indent level passed to json.dumps
        :param limit_reduction_script: if True, limits reduction_script to 50 chars in return
        :return: JSON dump of a dictionary representing the member variables
        """
        data_dict = attr.asdict(self)
        if limit_reduction_script:
            data_dict["reduction_script"] = data_dict["reduction_script"][:50]

        return json.dumps(data_dict, indent=indent)

    @staticmethod
    def deserialize(serialized_object):
        """
        Deserialize an object and return a dictionary of that object
        :param serialized_object: The object to deserialize
        :return: Dictionary of deserialized object
        """
        return json.loads(serialized_object)

    def populate(self, source, overwrite=True):
        """
        Populate the class from either a serialised object or a dictionary optionally retaining
        or overwriting existing values of attributes
        :param source: Object to populate class from
        :param overwrite: If True, overwrite existing values of attributes
        """
        if isinstance(source, str):
            try:
                source = self.deserialize(source)
            except json.decoder.JSONDecodeError as exp:
                raise ValueError(f"Unable to recognise serialized object {source}") from exp

        self_dict = attr.asdict(self)
        for key, value in source.items():
            if key in self_dict.keys():
                self_value = self_dict[key]
                if overwrite or self_value is None:
                    # Set the value of the variable on this object accessing it by name
                    setattr(self, key, value)
            else:
                raise ValueError(f"Unexpected key encountered during Message population: '{key}'.")

    def validate(self, destination):
        """
        Ensure that the message is valid to be sent to a given destination queue
        :param destination: (str) The name of the queue to send the data to
        """
        if destination == '/queue/DataReady':
            stages.validate_data_ready(self)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Exercise the functions that handle validating a message at each pipeline stage
"""
from unittest import TestCase, mock

from autoreduce_utils.message.validation import stages, validators
from autoreduce_utils.message.message import Message


class TestStages(TestCase):
    """
    Assert the stages return the correct bool value based on if they are valid
    """
    def setUp(self):
        self.valid_message = Message(
            run_number=1234,
            instrument='GEM',
            rb_number=1000000,
            started_by=-1,
            data='test/file/path',
        )
        self.invalid_message = Message(run_number='12345',
                                       instrument='not inst',
                                       rb_number=-1,
                                       started_by='test',
                                       data=123)

    def test_valid_validate_data_ready(self):
        """
        Test: validate_data_ready returns true
        When: supplied with a valid message
        """
        # This function raises on invalid message
        # if this test passes there was no raise
        stages.validate_data_ready(self.valid_message)

    def test_invalid_validate_data_ready(self):
        """
        Test: validate_data_ready returns false
        When: supplied with an invalid message
        """
        self.assertRaises(RuntimeError, stages.validate_data_ready, self.invalid_message)

    @mock.patch("autoreduce_utils.message.validation.stages.validators", spec=validators)
    def test_validate_data_ready_calls_validators(self, patched_validators):
        """
        Test: validate data calls the correct validators whilst running
        When: supplied with any message
        """
        stages.validate_data_ready(self.valid_message)

        patched_validators.validate_run_number.assert_called_once()
        patched_validators.validate_rb_number.assert_called_once()



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Validators to be used in the Message class
"""


def validate_run_number(run_number):
    """
    Assert a run number is valid
    :param run_number: The run number to validate
    """
    try:
        if int(run_number) > 0:
            return True
    except (ValueError, TypeError):
        return False
    return False


def validate_rb_number(rb_number):
    """
    Detects whether the RB number is a 7-digit number.

    :param rb_number:
    :return: False If the RB is not valid, otherwise true
    """
    try:
        rb_number = int(rb_number)
    except (ValueError, TypeError):
        return False

    return 999999 < rb_number < 10000000



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Functions that handle the processing of the validation dictionary
These can be considered helper functions to message.validation.stages
"""
import logging


def check_validity_dict(validity_dict):
    """
    Loop through a dictionary and assert if all are true
    If not log / print those that are false
    :param validity_dict: (dict) of key (name of validity check), values (result of check)
    :return: True if all values in dict are true, else false
    """
    is_valid = True
    for status in validity_dict.values():
        if not status:
            is_valid = False

    if is_valid:
        return True
    logging.error(dict_to_string(validity_dict))
    return False


def dict_to_string(validity_dict):
    """
    create and return a string representation of the validity check
    :param validity_dict: (dict) containing the validity checks
    :return: (str) A string to print
    """
    result = []
    for key, value in validity_dict.items():
        result.append(f"{key} - {value}")
    return ", ".join(result)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
# pylint: skip-file
import logging
import logging.handlers
from pathlib import Path

from autoreduce_utils.settings import CONFIG_ROOT

LOGGING_LEVEL = logging.WARNING
logging_dir = Path(CONFIG_ROOT, 'logs')
logging_dir.mkdir(parents=True, exist_ok=True)
LOGGING_LOC = logging_dir / 'isisicat_prefix_mappings.log'

logger = logging.getLogger('IsisICATPrefixMappings')
logger.setLevel(LOGGING_LEVEL)
handler = logging.handlers.RotatingFileHandler(LOGGING_LOC, maxBytes=104857600, backupCount=20)
handler.setLevel(LOGGING_LEVEL)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Generic client class used as an interface for other classes
"""

from abc import ABCMeta, abstractmethod

from autoreduce_utils.clients.settings.client_settings import ClientSettings


class AbstractClient:
    """
    Abstract class to ensure all Clients must have an init for credentials
    and connect, disconnect and _test_connection functions
    """
    __metaclass__ = ABCMeta

    credentials = None

    def __init__(self, credentials):
        if not isinstance(credentials, ClientSettings):
            raise TypeError("Expected instance of ClientSettings not {}".format(type(credentials)))
        self.credentials = credentials

    @abstractmethod
    def connect(self):
        """ Abstract function for connecting to a service """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def disconnect(self):
        """ Abstract function for disconnecting from a service """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _test_connection(self):
        """ Abstract function to test if a service connection has been made/is active """
        raise NotImplementedError  # pragma: no cover



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Test the Message class
"""
import unittest
import json

import attr

from autoreduce_utils.message.message import Message


class TestMessage(unittest.TestCase):
    """
    Test cases for the Message class used with AMQ
    """
    @staticmethod
    def _empty():
        """ Create and return an empty message object and corresponding dictionary"""
        empty_msg = Message()
        empty_dict = {
            'description': None,
            'facility': "ISIS",
            'run_number': None,
            'instrument': None,
            'rb_number': None,
            'started_by': None,
            'file_path': None,
            'overwrite': None,
            'run_version': None,
            'job_id': None,
            'reduction_script': None,
            'reduction_arguments': {},
            'reduction_log': "",
            'admin_log': "",
            'return_message': None,
            'retry_in': None,
            'software': None,
            'flat_output': False
        }
        return empty_msg, empty_dict

    @staticmethod
    def _populated():
        """ Create and return a populated message object and corresponding dictionary """
        run_number = 11111
        rb_number = 2222222
        description = 'test message'
        populated_msg = Message(run_number=run_number, rb_number=rb_number, description=description)
        populated_dict = {
            'description': description,
            'facility': "ISIS",
            'run_number': run_number,
            'instrument': None,
            'rb_number': rb_number,
            'started_by': None,
            'data': None,
            'overwrite': None,
            'run_version': None,
            'job_id': None,
            'reduction_script': None,
            'reduction_arguments': {},
            'reduction_log': "",
            'admin_log': "",
            'message': None,
            'retry_in': None,
            'reduction_data': None,
            'software': None,
            'flat_output': False
        }
        return populated_msg, populated_dict

    def test_init(self):
        """
        Test: All the expected member variables are created
        When: The class is initialised
        """
        empty_msg, _ = self._empty()
        self.assertEqual(empty_msg.description, "")
        self.assertEqual(empty_msg.facility, "ISIS")
        self.assertIsNone(empty_msg.run_number)
        self.assertIsNone(empty_msg.instrument)
        self.assertIsNone(empty_msg.rb_number)
        self.assertIsNone(empty_msg.started_by)
        self.assertIsNone(empty_msg.data)
        self.assertIsNone(empty_msg.overwrite)
        self.assertIsNone(empty_msg.run_version)
        self.assertIsNone(empty_msg.job_id)
        self.assertIsNone(empty_msg.reduction_script)
        self.assertEqual(empty_msg.reduction_arguments, {})
        self.assertEqual(empty_msg.reduction_log, "")
        self.assertEqual(empty_msg.admin_log, "")
        self.assertIsNone(empty_msg.message)
        self.assertIsNone(empty_msg.retry_in)
        self.assertIsNone(empty_msg.reduction_data)
        self.assertIsNone(empty_msg.software)
        self.assertFalse(empty_msg.flat_output)

    def test_to_dict_populated(self):
        """
        Test: A dictionary with populated values in produced
        When: attr.asdict() is called on a Message with populated values
        """
        populated_msg, populated_dict = self._populated()
        actual = attr.asdict(populated_msg)
        self.assertEqual(actual, populated_dict)

    def test_serialize_populated(self):
        """
        Test: An expected JSON object is produced
        When: The Message class is serialised
        """
        populated_msg, populated_dict = self._populated()
        serialized = populated_msg.serialize()
        actual = json.loads(serialized)
        self.assertEqual(actual, populated_dict)

    def test_deserialize_populated(self):
        """
        Test: A Dictionary with all the expected value is produced
        When: A populated serialized object is deserialized
        """
        populated_msg, populated_dict = self._populated()
        serialized = populated_msg.serialize()
        actual = populated_msg.deserialize(serialized)
        self.assertEqual(actual, populated_dict)

    def test_populate_from_dict_overwrite_true(self):
        """
        Test: Values are added and overwritten in the member variables
        When: Message.populate() is called with a non-empty dictionary and overwrite is True
        """
        _, populated_dict = self._populated()
        actual, _ = self._empty()
        actual.populate(source=populated_dict, overwrite=True)
        self.assertEqual(attr.asdict(actual), populated_dict)

    def test_populate_from_dict_overwrite_false(self):
        """
        Test: Values are added but NOT overwritten in the member variables
        When: Message.populate() is called with a non-empty dictionary and overwrite is False
        """
        populated_msg, populated_dict = self._populated()
        populated_dict['job_id'] = 123
        populated_dict['rb_number'] = 33333
        actual, _ = self._populated()
        actual.populate(source=populated_dict, overwrite=False)
        self.assertEqual(actual.rb_number, populated_msg.rb_number)
        self.assertEqual(actual.job_id, 123)

    def test_populate_from_serialized_overwrite_true(self):
        """
        Test: Values are added and overwritten in the member variables
        When: Message.populate() called with a non-empty serialized object and overwrite is True
        """
        populated_msg, populated_dict = self._populated()
        serialized = populated_msg.serialize()
        actual, _ = self._empty()
        actual.populate(source=serialized, overwrite=True)
        self.assertEqual(attr.asdict(actual), populated_dict)

    def test_populate_from_serialized_overwrite_false(self):
        """
        Test: Values are added but NOT overwritten in the member variables
        When: Message.populate() called with a non-empty serialized object and overwrite is False
        """
        new_msg, _ = self._populated()
        new_msg.job_id = 123
        new_msg.rb_number = 33333
        serialized = new_msg.serialize()
        actual, _ = self._populated()
        original, _ = self._populated()
        actual.populate(source=serialized, overwrite=False)
        self.assertEqual(actual.rb_number, original.rb_number)
        self.assertEqual(actual.job_id, 123)

    def test_invalid_serialized(self):
        """
        Test: A ValueError is raised
        When: An invalid serialized object is supplied to Message.populate()
        """
        serialized = 'test'
        empty_msg, _ = self._empty()
        self.assertRaises(ValueError, empty_msg.populate, serialized)

    def test_populate_with_invalid_key(self):
        """
        Test: A warning is logged
        When: An unknown key is used to populate the Message
        """
        args = {'unknown': True}
        msg = Message()
        with self.assertRaises(ValueError):
            msg.populate(args)

    def test_validate_data_ready_valid(self):
        """
        Test: No exception is raised
        When: Calling validate for data_ready with a valid message
        """
        message = Message(instrument='GEM',
                          run_number=111,
                          rb_number=2222222,
                          data='file/path',
                          facility="ISIS",
                          started_by=0)
        try:
            self.assertIsNone(message.validate('/queue/DataReady'))
        except RuntimeError:
            self.fail()

    def test_validate_data_ready_invalid(self):
        """
        Test: An exception is raised
        When: An invalid Message is validated
        """
        message = Message(instrument='Not an inst')
        self.assertRaises(RuntimeError, message.validate, '/queue/DataReady')



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Exercise the helper functions that process the validation results
"""
import unittest

from unittest.mock import patch
from autoreduce_utils.message.validation import process


class TestProcess(unittest.TestCase):
    def setUp(self):
        self.true_validity_dict = {
            'check_1': True,
            'check_2': True,
        }
        self.mixed_validity_dict = {
            'check_1': True,
            'check_2': False,
        }
        self.false_validity_dict = {
            'check_1': False,
            'check_2': False,
        }

    def test_validity_dict_valid(self):
        """
        Test: The return values
        When: Different dicts are supplied to check_validity_dict (all true, mixed, all false)
        """
        self.assertTrue(process.check_validity_dict(self.true_validity_dict))
        self.assertFalse(process.check_validity_dict(self.mixed_validity_dict))
        self.assertFalse(process.check_validity_dict(self.false_validity_dict))

    @patch('autoreduce_utils.message.validation.process.dict_to_string')
    @patch('logging.error')
    def test_validity_dict_log_when_false(self, mock_err_log, mock_dict_to_str):
        """
        Test: A logging error is called with output of dict_to_string
        When: check_validity_dict returns false
        """
        msg = 'Test string'
        mock_dict_to_str.return_value = msg
        self.assertFalse(process.check_validity_dict(self.false_validity_dict))
        mock_err_log.assert_called_once_with(msg)

    def test_dict_to_string(self):
        """
        Test: An expected string is returned
        When: dict_to_string is called with a valid dict
        """
        expected = "check_1 - True, check_2 - True"
        self.assertEqual(process.dict_to_string(self.true_validity_dict), expected)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Tests for the ClientSettingsFactory
"""
import unittest

from autoreduce_utils.clients.settings.client_settings_factory import (ClientSettingsFactory, MySQLSettings,
                                                                       ICATSettings, ActiveMQSettings)


# pylint:disable=missing-docstring
class TestClientSettingsFactory(unittest.TestCase):
    def setUp(self):
        self.factory = ClientSettingsFactory()

    def test_create_database(self):
        actual = self.factory.create('database',
                                     username='test-user',
                                     password='test-pass',
                                     host='test-host',
                                     port='test-port',
                                     database_name='test-name')
        self.assertIsInstance(actual, MySQLSettings)
        self.assertEqual(actual.username, 'test-user')
        self.assertEqual(actual.password, 'test-pass')
        self.assertEqual(actual.host, 'test-host')
        self.assertEqual(actual.port, 'test-port')
        self.assertEqual(actual.database, 'test-name')
        self.assertEqual(actual.get_full_connection_string(), 'mysql+mysqldb://test-user:test-pass@test-host/test-name')

    def test_create_queue(self):
        actual = self.factory.create('queue',
                                     username='test-user',
                                     password='test-pass',
                                     host='test-host',
                                     port='test-port',
                                     data_ready='test-dr')
        self.assertIsInstance(actual, ActiveMQSettings)
        self.assertEqual(actual.username, 'test-user')
        self.assertEqual(actual.password, 'test-pass')
        self.assertEqual(actual.host, 'test-host')
        self.assertEqual(actual.port, 'test-port')
        self.assertEqual(actual.data_ready, 'test-dr')

    def test_create_icat(self):
        actual = self.factory.create('icat',
                                     username='test-user',
                                     password='test-pass',
                                     host='test-host',
                                     port='test-port',
                                     authentication_type='test-auth')
        self.assertIsInstance(actual, ICATSettings)
        self.assertEqual(actual.username, 'test-user')
        self.assertEqual(actual.password, 'test-pass')
        self.assertEqual(actual.host, 'test-host')
        self.assertEqual(actual.port, 'test-port')
        self.assertEqual(actual.auth, 'test-auth')

    def test_invalid_not_a_factory(self):
        self.assertRaisesRegex(
            ValueError, f"Factories creation settings type must be one of:"
            f"{','.join(self.factory.valid_types)}", self.factory.create, 'not-factory', 'user', 'pass', 'host', 'port')

    def test_invalid_database_args(self):
        self.assertRaisesRegex(ValueError, "database_invalid is not a recognised key "
                               "word argument.",
                               self.factory.create,
                               'database',
                               'user',
                               'pass',
                               'host',
                               'port',
                               database_invalid='invalid')

    def test_invalid_queue_args(self):
        self.assertRaisesRegex(ValueError,
                               "queue_invalid is not a recognised key word argument.",
                               self.factory.create,
                               'queue',
                               'user',
                               'pass',
                               'host',
                               'port',
                               queue_invalid='invalid')

    def test_invalid_icat_args(self):
        self.assertRaisesRegex(ValueError,
                               "icat_invalid is not a recognised key word argument.",
                               self.factory.create,
                               'icat',
                               'user',
                               'pass',
                               'host',
                               'port',
                               icat_invalid='invalid')



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module to perform ICAT client functionality
Functions for login and query available from class
"""

import icat

from autoreduce_utils.settings import ICAT_SETTINGS
from autoreduce_utils.clients.abstract_client import AbstractClient
from autoreduce_utils.clients.connection_exception import ConnectionException


class ICATClient(AbstractClient):
    """
    This class provides a layer of abstraction from Python ICAT.
    Only allowing logging in and querying.
    """
    def __init__(self, credentials=None):
        if not credentials:
            credentials = ICAT_SETTINGS
        super(ICATClient, self).__init__(credentials)  # pylint:disable=super-with-arguments
        self.client = icat.Client(self.credentials.host)

    def connect(self):
        """
        Log in to ICAT using the details provided in the credentials.ini file
        """
        self.client.login(auth=self.credentials.auth,
                          credentials={
                              'username': self.credentials.username,
                              'password': self.credentials.password
                          })

    def _test_connection(self):
        """
        Test that the connection has been successful
        """
        try:
            self.client.refresh()
        except icat.exception.ICATSessionError as exp:
            raise ConnectionException("ICAT") from exp
        return True

    def refresh(self):
        """ Refreshes the ICAT session only if necessary """
        self.client.refresh()

    def disconnect(self):
        """ Disconnect the ICAT client """
        self.client.logout()

    def execute_query(self, query):
        """
        Runs a query on ICAT - assumes a valid login has already been obtained
        :param query: The query to run
        :return: The result of the query
        """
        try:
            self.client.refresh()
        except icat.exception.ICATSessionError:
            # Session has most likely expired, try and log in again.
            # Have to set sessionId to None otherwise python ICAT attempts
            # to log out with an expired sessionId
            self.client.sessionId = None
            self.connect()
        return self.client.search(query)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Exercise the validation functions
"""
import unittest

from autoreduce_utils.message.validation import validators


class TestValidators(unittest.TestCase):
    """
    As the validators are simple bool returns,
    we are validating a function completely per test case
    """
    def test_validate_run_number(self):
        """
        Test: validate_run_number returns the expected result
        When: In valid and invalid cases
        """
        self.assertTrue(validators.validate_run_number(1))
        self.assertTrue(validators.validate_run_number('001'))

        self.assertFalse(validators.validate_run_number(0))
        self.assertFalse(validators.validate_run_number(-1))
        self.assertFalse(validators.validate_run_number('string'))

    def test_validate_rb_number(self):
        """
        Tests the valid and invalid cases for RB numbers
        """
        valid_values = [1000000, 2000000, "1000000", 9999999]
        for i in valid_values:
            self.assertTrue(validators.validate_rb_number(i), "Failed with value %s" % i)

        invalid_values = [0, 0.1, -1, -100, None, "foo", 12345678910, "1231435252242"]
        for i in invalid_values:
            self.assertFalse(validators.validate_rb_number(i), "Failed with value %s" % i)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Test functionality for the activemq client
"""
import os
import socket

from unittest import TestCase, mock
from unittest.mock import patch, MagicMock

from autoreduce_utils.message.message import Message
from autoreduce_utils.clients.connection_exception import ConnectionException
from autoreduce_utils.clients.queue_client import QueueClient
from autoreduce_utils.clients.settings.client_settings_factory import ClientSettingsFactory
from autoreduce_utils.settings import ACTIVEMQ_SETTINGS


# pylint:disable=protected-access,no-self-use
class TestQueueClient(TestCase):
    """
    Exercises the queue client
    """
    def test_default_init(self):
        """
        Test: Class variables are created and set
        When: QueueClient is initialised with default credentials
        """
        client = QueueClient()
        self.assertIsNotNone(client.credentials)
        self.assertIsNone(client._connection)
        self.assertEqual('queue_client', client._consumer_name)

    def test_invalid_init(self):
        """
        Test: A TypeError is raised
        When: QueueClient is initialised with invalid credentials
        """
        self.assertRaises(TypeError, QueueClient, 'string')

    def test_valid_connection(self):
        """
        Test: Access is established with a valid connection
        (This by proxy will also test the get_connection function)
        When: connect is called while valid credentials are held
        """
        client = QueueClient()
        client.connect()
        self.assertTrue(client._connection.is_connected())

    def test_connection_failed_invalid_credentials(self):
        """
        Test: A ConnectionException is raised
        When: _test_connection is called while invalid credentials are held
        """
        incorrect_credentials = ClientSettingsFactory().create('queue',
                                                               username='not-user',
                                                               password='not-pass',
                                                               host='127.does.not.exist',
                                                               port='1234')
        client = QueueClient(incorrect_credentials)
        with self.assertRaises(ConnectionException):
            client.connect()

    def test_stop_connection(self):
        """
        Test: Connection is stopped and connection variables are set to None
        When: disconnect is called while a valid connection is currently established
        """
        client = QueueClient()
        mocked_connection = mock.Mock()
        client._connection = mocked_connection

        with mock.patch("uuid.uuid4") as patched_uuid:
            patched_uuid.return_value = 1
            client.disconnect()

        mocked_connection.disconnect.assert_called_with(receipt=str(1))
        self.assertIsNone(client._connection)

    @patch('stomp.connect.StompConnection11.send')
    def test_send_with_raw_string(self, mock_stomp_send):
        """
        Test: send sends the given data using stomp.send
        When: send is called with a string argument for message
        """
        client = QueueClient()
        client.send('dataready', 'raw_json_dump')
        (args, _) = mock_stomp_send.call_args
        self.assertEqual(args[0], 'dataready')
        self.assertEqual(args[1], 'raw_json_dump')

    @patch('stomp.connect.StompConnection11.send')
    def test_send_with_message_instance(self, mock_stomp_send):
        """
        Test: send sends the given data using stomp.send
        When: send is called with a Message instance argument for message
        """
        client = QueueClient()
        message = Message(description="test-message")
        client.send('dataready', message)
        (args, _) = mock_stomp_send.call_args
        self.assertEqual(args[0], 'dataready')
        self.assertEqual(args[1], message.serialize())

    @patch('stomp.connect.StompConnection11.ack')
    def test_ack(self, mock_stomp_ack):
        """
        Test: ack sends an ack frame using stomp.ack
        When: ack is called while a valid connection is held
        """
        client = QueueClient()
        client.connect()
        client.ack("test", "subscription")
        mock_stomp_ack.assert_called_once_with('test', "subscription")

    def test_create_connection_bad_development(self):
        """
        Test: Exception raised
        When: production host used in non production environment
        """
        client = QueueClient()
        real_host = client.credentials.host
        client.credentials.host = "production.domain.com"
        self.assertRaisesRegex(RuntimeError, "non-development", client._create_connection)
        client.credentials.host = real_host

    def test_create_connection_bad_production(self):
        """
        Test: Exception raised
        When: Local host used in production environment
        """
        client = QueueClient()
        real_host = client.credentials.host

        os.environ["AUTOREDUCTION_PRODUCTION"] = "1"
        client.credentials.host = "127.0.0.1"
        self.assertRaisesRegex(RuntimeError, ".*production environment.*", client._create_connection)

        client.credentials.host = "somethingdev"
        self.assertRaisesRegex(RuntimeError, ".*production environment.*", client._create_connection)

        client.credentials.host = "activemq"
        self.assertRaisesRegex(RuntimeError, ".*production environment.*", client._create_connection)

        client.credentials.host = real_host
        del os.environ["AUTOREDUCTION_PRODUCTION"]

    def test_test_connection_not_connected(self):
        """
        Test: Exception raised
        When: test_connection called when not connected

        """
        client = QueueClient()
        mock_connection = MagicMock()
        mock_connection.is_connected.return_value = False
        client._connection = mock_connection
        with self.assertRaises(ConnectionException):
            client._test_connection()

    def test_test_connection_connected(self):
        """
        Test: test_connection returns True
        When: Connected
        """
        client = QueueClient()
        mock_connection = MagicMock()
        mock_connection.is_connected.return_value = True
        client._connection = mock_connection
        self.assertTrue(client._test_connection())

    def test_subscribe(self):
        """
        Test: correct calls made
        When: subscribe is called
        """
        client = QueueClient()
        mock_connection = MagicMock()
        mock_listener = MagicMock()
        client._connection = mock_connection
        client.subscribe(mock_listener)

        mock_connection.set_listener.assert_called_with("queue_processor", mock_listener)
        mock_connection.subscribe.assert_called_with(destination=ACTIVEMQ_SETTINGS.data_ready,
                                                     id=socket.getfqdn(),
                                                     ack="client-individual",
                                                     header={"activemq.prefetchSize": "1"})



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Generic settings class for client access
"""


class ClientSettings:
    """
    Hold common values for all Settings object
    """

    username = None
    password = None
    host = None
    port = None

    def __init__(self, username, password, host, port):
        self.username = self._attempt_param_cast(username)
        self.password = self._attempt_param_cast(password)
        self.host = self._attempt_param_cast(host)
        self.port = self._attempt_param_cast(port)

    @staticmethod
    def _attempt_param_cast(param):
        """
        Raise exception if param is not string castable
        """
        try:
            # Bool and float values are clearly easy mistakes
            if isinstance(param, (bool, float)):
                raise TypeError()

            # Python: Easier to ask for forgiveness than permission
            # So just attempt a cast to str, don't check if we have a string
            # so we can use mock objects in-lieu
            return str(param)
        except TypeError as exp:
            raise ValueError(f"{param} of type {type(param)} is not a string") from exp



# pylint:skip-file
"""
Wrapper for the functionality for various installation and project setup commands
see:
    `python setup.py help`
for more details
"""
from os import path
from setuptools import setup, find_packages

# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='autoreduce_utils',
      version='0.1.0',
      description='ISIS Autoreduce',
      author='ISIS Autoreduction Team',
      url='https://github.com/ISISScientificComputing/autoreduce-utils/',
      install_requires=[
          'Django==3.2',
          'gitpython==3.1.14',
          'python-icat==0.18.1',
          'suds-py3==1.4.4.1',
          'stomp.py==6.1.0',
      ],
      packages=find_packages(),
      package_data={"autoreduce_utils": ["test_credentials.ini"]},
      entry_points={"console_scripts": ["autoreduce-creds-migrate = autoreduce_utils.migrate_credentials:main"]},
      long_description=long_description,
      long_description_content_type='text/markdown')



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Custom class for connection errors
"""


class ConnectionException(Exception):
    """
    Simple class for raising exceptions when we cannot connect to services
    """
    def __init__(self, service_name):
        message = "Unable to connect to {0} with provided credentials. " \
                  "Please check the {0} settings files then try again.".format(service_name)
        super(ConnectionException, self).__init__(message)  # pylint:disable=super-with-arguments



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
get_icat_instrument_prefix() can be used to map Autoreduction to ICAT instrument prefixes
"""
from autoreduce_utils.clients.icat_client import ICATClient
from autoreduce_utils.clients.tools.isisicat_prefix_mapping_logging_setup import logger


def get_icat_instrument_prefix(instrument_fullname: str) -> str:
    """
    Queries ICAT for shorter names for all Autoreduction instruments or only selection if passed in
    :return: instrument prefix
    """
    client = ICATClient()

    try:
        icat_instruments = client.execute_query("SELECT i FROM Instrument i")
    except Exception as exc:
        warning_message = "ICAT instrument query failed"
        print(warning_message)
        logger.warning(warning_message)
        raise RuntimeError(warning_message) from exc

    icat_instrument = next((x for x in icat_instruments if x.fullName == instrument_fullname), None)

    if not icat_instrument:
        warning_message = f"No instrument in ICAT with fullName {instrument_fullname}"
        print(warning_message)
        logger.warning(warning_message)
        raise RuntimeError(f"Instrument with fullname {instrument_fullname} not found in ICAT.")

    return icat_instrument.name



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Test ICAT prefix mapping
"""
import unittest
from unittest.mock import patch

from autoreduce_utils.clients.tools.isisicat_prefix_mapping import get_icat_instrument_prefix


# pylint:disable=no-self-use
class MockInstrumentQueryResult:
    """
    Mocks result of isisicat_prefix_mapping.client.execute_query for an instrument
    """
    def __init__(self, name, full_name):
        self.name = name
        # camelCase used due to mocking of external class
        self.fullName = full_name  # pylint:disable=invalid-name


class TestICATPrefixMapping(unittest.TestCase):
    """
    Test ICAT prefix mapping
    """
    DIR = "utils.clients.tools.isisicat_prefix_mapping"

    @patch('icat.Client.__init__', return_value=None)
    @patch('autoreduce_utils.clients.icat_client.ICATClient.execute_query',
           return_value=[MockInstrumentQueryResult("ENG", "ENGINX")])
    def test_get_icat_instrument_prefix_executes_icat_query(self, mock_execute_query, _):
        """
        Test: If instruments are fetched through a query
        When: Testing if get_icat_instrument_prefix executes an ICAT query
        """
        self.assertEqual("ENG", get_icat_instrument_prefix("ENGINX"))
        mock_execute_query.assert_called_once()

    @patch('icat.Client.__init__', return_value=None)
    @patch('autoreduce_utils.clients.icat_client.ICATClient.execute_query', side_effect=Exception)
    def test_get_icat_instrument_prefix_icat_query_raises(self, mock_execute_query, _):
        """
        Test: If the query raises due to a connection issue or wrong credentials
        When: Testing if get_icat_instrument_prefix executes an ICAT query
        """
        self.assertRaises(RuntimeError, get_icat_instrument_prefix, "ENGINX")
        mock_execute_query.assert_called_once()

    @patch('icat.Client.__init__', return_value=None)
    @patch('logging.Logger.warning')
    @patch('autoreduce_utils.clients.icat_client.ICATClient.execute_query',
           return_value=[MockInstrumentQueryResult("ENG", "ENGINX")])
    def test_get_icat_instrument_prefix_log_invalid_instrument(self, _mock_execute_query, mock_logger_warning, _):
        """
        Test: RuntimeError raised
        When: get_instrument_prefix called on invalid instrument
        """
        self.assertRaises(RuntimeError, get_icat_instrument_prefix, "INVALIDINSTRUMENT")
        mock_logger_warning.assert_called_once()



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Combination of validation checks to be performed on the Message at each stage of the pipeline
"""
import autoreduce_utils.message.validation.validators as validators
from autoreduce_utils.message.validation.process import dict_to_string


def validate_data_ready(message):
    """
    Assert a message is ready to be passed to the /DataReady queue
    :param message: A message object to be validated
    :return: True if valid
    """
    validity_dict = {
        'run_number': validators.validate_run_number(message.run_number),
        'rb_number': validators.validate_rb_number(message.rb_number),
        'started_by': isinstance(message.started_by, int),
        'file_path': isinstance(message.data, str),
        'facility': isinstance(message.facility, str)
    }
    if False in validity_dict.values():
        raise RuntimeError(f"Validation failed: {dict_to_string(validity_dict)}")



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Factory for creating settings objects that can be used in the client classes
"""
# pylint: disable=fixme
from autoreduce_utils.clients.settings.client_settings import ClientSettings


class ClientSettingsFactory:
    """
    Class for the settings factory
    """

    ignore_kwargs = ['username', 'password', 'host', 'port']
    valid_types = ['database', 'icat', 'queue', 'sftp', 'cycle']

    def create(self, settings_type, username, password, host, port, **kwargs):
        """
        Create a settings object to use with a client
        :param settings_type: The type of client you require this can be: database, icat or queue
        :param username: username for logon
        :param password: password for logon
        :param host: host address for service
        :param port: port on the host machine where the service is running
        :param kwargs: key word arguments used for specific classes
                       Database specific args : database_name
                       ICAT specific args     : authentication_type
                       Queue specific args    : reduction_pending, data_ready, reduction_started
                                                reduction_complete, reduction_error
        :return: A ClientSettings object
        """
        if settings_type.lower() not in self.valid_types:
            raise ValueError(f"Factories creation settings type must be one of:" f"{','.join(self.valid_types)}")
        kwargs['username'] = username
        kwargs['password'] = password
        kwargs['host'] = host
        kwargs['port'] = port

        settings = None
        if settings_type.lower() == 'database':
            settings = self._create_database(**kwargs)
        elif settings_type.lower() == 'icat':
            settings = self._create_icat(**kwargs)
        elif settings_type.lower() == 'queue':
            settings = self._create_queue(**kwargs)
        elif settings_type.lower() == 'sftp':
            settings = self._create_sftp(**kwargs)
        elif settings_type.lower() == 'cycle':
            settings = self._create_cycle(**kwargs)
        return settings

    def _create_database(self, **kwargs):
        """
        :return: Database compatible settings object
        """
        database_kwargs = ['database_name']
        self._test_kwargs(database_kwargs, kwargs)
        return MySQLSettings(**kwargs)

    def _create_queue(self, **kwargs):
        """
        :return: Queue compatible settings object
        """
        queue_kwargs = ['data_ready']
        self._test_kwargs(queue_kwargs, kwargs)
        return ActiveMQSettings(**kwargs)

    def _create_icat(self, **kwargs):
        """
        :return: Icat compatible settings object
        """
        icat_kwargs = ['authentication_type']
        self._test_kwargs(icat_kwargs, kwargs)
        return ICATSettings(**kwargs)

    def _create_sftp(self, **kwargs):
        """
        :return: SFTP compatible settings object
        """
        sftp_kwargs = []  # No additional kwargs needed for sftp
        self._test_kwargs(sftp_kwargs, kwargs)
        return SFTPSettings(**kwargs)

    def _create_cycle(self, **kwargs):
        """
        :return: cycle-ingestion compatible settings object
        """
        cycle_kwargs = ['uows_url', 'scheduler_url']
        self._test_kwargs(cycle_kwargs, kwargs)
        return CycleIngestionSettings(**kwargs)

    def _test_kwargs(self, expected, actual):
        """
        Ensure that the kwargs given as input contain the expected keys
        """
        for key, _ in actual.items():
            if key not in expected and key not in self.ignore_kwargs:
                raise ValueError("{0} is not a recognised key word argument."
                                 " Valid kwargs: {1}".format(key, expected))


class ICATSettings(ClientSettings):
    """
    ICAT settings object
    """
    auth = None

    def __init__(self, authentication_type='Simple', **kwargs):
        super(ICATSettings, self).__init__(**kwargs)  # pylint:disable=super-with-arguments
        self.auth = authentication_type


class MySQLSettings(ClientSettings):
    """
    MySQL settings to be used as a Database settings object
    """
    database = None

    def __init__(self, database_name='autoreduction', **kwargs):
        super(MySQLSettings, self).__init__(**kwargs)  # pylint:disable=super-with-arguments
        self.database = database_name

    def get_full_connection_string(self):
        """ :return: string for connecting directly to mysql service with user + pass """
        return 'mysql+mysqldb://{0}:{1}@{2}/{3}'.format(self.username, self.password, self.host, self.database)


class ActiveMQSettings(ClientSettings):
    """
    ActiveMq settings to be used as a Queue settings object
    """
    data_ready = None
    all_subscriptions = None

    def __init__(self, data_ready='/queue/DataReady', **kwargs):
        # TODO explicitly state args
        super(ActiveMQSettings, self).__init__(**kwargs)  # pylint:disable=super-with-arguments

        self.data_ready = data_ready
        self.all_subscriptions = [data_ready]


class SFTPSettings(ClientSettings):
    """
    SFTP settings object
    """
    def __init__(self, **kwargs):  # pylint:disable=useless-super-delegation
        super(SFTPSettings, self).__init__(**kwargs)  # pylint:disable=super-with-arguments


class CycleIngestionSettings(ClientSettings):
    """
    Cycle-ingestion settings object
    """
    uows_url = None
    scheduler_url = None

    # pylint:disable=super-with-arguments
    def __init__(self, uows_url, scheduler_url, **kwargs):
        super(CycleIngestionSettings, self).__init__(**kwargs)
        self.uows_url = uows_url
        self.scheduler_url = scheduler_url



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Test AbstractClient functionality
"""
import unittest

from autoreduce_utils.clients.abstract_client import AbstractClient
from autoreduce_utils.clients.settings.client_settings import ClientSettings


# pylint:disable=missing-docstring
class TestAbstractClient(unittest.TestCase):
    def test_client_settings_init(self):
        """
        Test: A client is created
        When: A class which implements AbstractClient is initiated with a ClientSettings argument
        """
        valid_settings = ClientSettings(username='user', password='pass', host='host', port='123')
        interface = ClientWrapper(credentials=valid_settings)
        self.assertIsNotNone(interface)

    def test_derived_settings_init(self):
        """
        Test: A client is created
        When: A class which implements AbstractClient is initiated with an argument that
        extends ClientSettings
        """
        class DerivedSettings(ClientSettings):
            pass

        derived_settings = DerivedSettings(username='user', password='pass', host='host', port='123')

        interface = ClientWrapper(credentials=derived_settings)
        self.assertIsNotNone(interface)

    def test_invalid_init(self):
        self.assertRaisesRegex(TypeError, "Expected instance of ClientSettings not <class 'int'>", ClientWrapper, 10)
        self.assertRaisesRegex(TypeError, "Expected instance of ClientSettings not <class 'str'>", ClientWrapper,
                               'string')
        self.assertRaisesRegex(TypeError, "Expected instance of ClientSettings not <class 'list'>", ClientWrapper,
                               [1, 2, 3, 4])


class ClientWrapper(AbstractClient):  # pragma: no cover
    """
    Class that implements the AbstractClient.
    This is used for testing AbstractClient.__init__() - specifically, attempting to supply
    invalid ClientSettings objects.
    See use in utils.clients.tests..test_abstract_client.test_invalid_init (above)
    """
    def connect(self):
        pass

    def disconnect(self):
        pass

    def _test_connection(self):
        pass
